package net.sergeych.kiloparsec

import kotlinx.coroutines.CancellationException
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.launch
import net.sergeych.crypto2.SigningKey
import net.sergeych.kiloparsec.adapter.InetTransportDevice
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.debug
import net.sergeych.mp_logger.exception
import net.sergeych.mp_logger.info
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.tools.AtomicCounter

private val instances = AtomicCounter()

/**
 * The Kiloparsec server.
 * Server accepts incoming connections and serves them using the same [clientInterface].
 *
 * ## Incoming connections
 *
 * Server collecting incoming connections provided by [connections] `Flow`. For each incoming connection
 * the Kiloparsec handshake is performed, then the session object is created, see below, and connection is
 * served with [clientInterface] until closed.
 *
 * ## Session param [S]
 *
 * After the successful handshake server creates new session for each connection calling the [sessionBuilder].
 * Then it creates a [KiloScope] so [KiloScope.session] holds this connection-specific instance. Then
 * [KiloInterface.onConnected] is called with this scope. Session can be used to hold connection state. Session
 * objects are not persistent, but could be initialized in [KiloInterface.onConnected] where the remote side
 * [KiloScope.remoteIdentity] is already verified and set.
 *
 * ## Usage:
 *
 * Create a shared library between you server and clients, to specify the interface (otherwise you can
 * share sources).
 *
 * Suppose we have session with a state:
 * ```kotlin
 * data class Session(
 *     var data: String,
 * )
 *```
 *
 * And some commands to access and change it, in the shared library too:
 *
 * ```kotlin
 * val cmdSave by command<String, Unit>()
 * val cmdLoad by command<Unit, String>()
 * val cmdDrop by command<Unit, Unit>()
 * val cmdException by command<Unit, Unit>()
 * ```
 * Then the server code (TCP/IP variant) could look like:
 *
 * ```kotlin
 * // The server implementation (could be shared between server instances connected
 * // to different protocol adapters):
 *
 * val cli = KiloInterface<Session>().apply {
 *     // Suppose we want to throw this exception at the caller site, so we need to register it:
 *     registerError { SomeException() }
 *
 *     // Session initialization. If you need a sessino to depend initially on the client's identity.
 *     // you can do it here:
 *     onConnected {
 *         // check the remoteIdentity
 *         session.data = if( remoteIdentity == somePublicVerifyingKey )
 *              "known"
 *         else
 *              "unknown"
 *     }
 *     on(cmdSave) { session.data = it }
 *     on(cmdLoad) {
 *         session.data
 *     }
 *     on(cmdException) {
 *         throw TestException()
 *     }
 *     on(cmdDrop) {
 *         throw LocalInterface.BreakConnectionException()
 *     }
 * }
 *
 * // Create the server instance that accepts incoming TCP/IP connections on all local interfaces using the
 * // specified port:
 *
 * val server = KiloServer(cli, acceptTcpDevice(port)) {
 *     // This creates a new session
 *     Session("unknown")
 * }
 * ```
 *
 * See [KiloClient] to connect to the server.
 *
 * @param S the type of the server session object, returned by [sessionBuilder]. See above
 * @param clientInterface the interface available for remote calls
 * @param connections flow of incoming connections. Server stops when the flow is fully collected (normally
 *          it shouldn't)
 * @param serverSecretKey the [SigningKey] used to identify this server during Kiloparsec handshake.
 * @param sessionBuilder callback that creates session objects for successful incoming connections
 */
class KiloServer<S>(
    private val clientInterface: KiloInterface<S>,
    private val connections: Flow<InetTransportDevice>,
    private val serverSecretKey: SigningKey? = null,
    private val sessionBuilder: () -> S,
) : LogTag("KS:${instances.incrementAndGet()}") {

    private val job = globalLaunch {
        connections.collect { device ->
            launch {
                try {
                    info { "connected ${device}" }
                    KiloServerConnection(clientInterface, device, sessionBuilder(), serverSecretKey)
                        .apply { debug { "server connection is ready" } }
                        .run()
                } catch (_: CancellationException) {
                } catch (cce: LocalInterface.BreakConnectionException) {
                    info { "Closed exception caught, closing (${cce.flushSendQueue}" }
                } catch (t: Throwable) {
                    exception { "unexpected while creating kiloclient" to t }
                }
            }
        }
    }

    /**
     * Stop the server and cancel all pending sessions. Unlike finishing the flow passed
     * for [KiloServer.connections], it will cancel all currently active sessions.
     */
    fun close() {
        job.cancel()
    }

    /**
     * Server is closed either if [close] was called or all [KiloServer.connections] were collected and flow was
     * closed.
     */
    @Suppress("unused")
    val isClosed: Boolean get() = job.isCompleted
}