package net.sergeych.kiloparsec.adapter

import io.ktor.network.selector.*
import io.ktor.network.sockets.*
import io.ktor.utils.io.*
import kotlinx.coroutines.*
import kotlinx.coroutines.CancellationException
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.ClosedReceiveChannelException
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.flow
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import kotlinx.datetime.Clock
import net.sergeych.kiloparsec.AsyncVarint
import net.sergeych.kiloparsec.KiloClient
import net.sergeych.kiloparsec.KiloServer
import net.sergeych.kiloparsec.LocalInterface
import net.sergeych.mp_logger.*
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.tools.AtomicCounter
import net.sergeych.tools.AtomicValue
import kotlin.time.Duration.Companion.seconds

private val logCounter = AtomicCounter(0)

class ProtocolException(text: String, cause: Throwable? = null) : RuntimeException(text, cause)

const val MAX_TCP_BLOCK_SIZE = 16776216
internal val PING_INACTIVITY_TIME = 30.seconds

/**
 * Listen for incoming TCP/IP connections on all local interfaces and the specified [port]
 * anc create flow of [InetTransportDevice] suitable for [KiloClient].
 */
fun acceptTcpDevice(port: Int,localInterface: String = "0.0.0.0"): Flow<InetTransportDevice> {
    val selectorManager = SelectorManager(Dispatchers.IO)
    val serverSocket = aSocket(selectorManager).tcp().bind(localInterface, port)
    return flow {
        while (true) {
            serverSocket.accept().let { sock ->
                emit(inetTransportDevice(sock, "srv"))
            }
        }
    }
}

suspend fun connectTcpDevice(address: String) = connectTcpDevice(address.toNetworkAddress())

/**
 * Connect to the TCP/IP server (see [KiloServer]) at the specified address and provide th compatible
 * [InetTransportDevice] to use with [KiloClient].
 */
suspend fun connectTcpDevice(address: NetworkAddress): InetTransportDevice {
    val selectorManager = SelectorManager(Dispatchers.IO)
    val socket = aSocket(selectorManager).tcp().connect(address.host, address.port)
    return inetTransportDevice(socket)
}

/**
 * Parse `host:port` string into the [NetworkAddress]
 */
fun String.toNetworkAddress(): NetworkAddress {
    val (host, port) = this.split(":").map { it.trim() }
    return NetworkAddress(host, port.toInt())
}

private fun inetTransportDevice(
    sock: Socket,
    suffix: String = "cli",
): InetTransportDevice {
    val networkAddress = (sock.remoteAddress as InetSocketAddress).let { NetworkAddress(it.hostname, it.port) }
    val inputBlocks = Channel<UByteArray>(4096)
    val outputBlocks = Channel<UByteArray>(4096)

    val log = LogTag("TCPT${logCounter.incrementAndGet()}:$suffix:$networkAddress")
    val job = AtomicValue<Job?>(null)

    val sockOutput = sock.openWriteChannel()
    val sockInput = runCatching { sock.openReadChannel() }.getOrElse {
        log.warning { "failed to open read channel $it" }
        throw IllegalStateException("failed to open read channel")
    }

    fun stop() {
        job.mutate {
            if ( it != null ) {
                log.debug { "stopping" }
                runCatching { inputBlocks.close() }
                runCatching { outputBlocks.close() }
                // The problem: on mac platofrms closing the socket does not close its input
                // and output channels!
                runCatching { sockInput.cancel() }
                runCatching { sockOutput.close() }
                if (!sock.isClosed)
                    runCatching {
                        log.debug { "closing socket by stop" }
                        sock.close()
                    }
                else
                    log.debug { "socket is already closed when stop is called" }
                it.cancel()
                log.debug { "implementation job cancel called" }
            } else
                log.debug { "already stopped" }
            null
        }
    }

    var lastActiveAt = Clock.System.now()
    job.value = globalLaunch {
        launch {

            log.debug { "opening read channel" }

            while (isActive && sock.isActive) {
                try {
                    val size = AsyncVarint.decodeUnsigned(sockInput).toInt()
                    if (size > MAX_TCP_BLOCK_SIZE) // 16M is a max command block
                        throw ProtocolException("Illegal block size: $size should be < $MAX_TCP_BLOCK_SIZE")
                    val data = ByteArray(size)
                    if (size == 0) {
                        log.debug { "ping received" }
                        lastActiveAt = Clock.System.now()
                    } else {
                        sockInput.readFully(data, 0, size)
                        inputBlocks.send(data.toUByteArray())
                    }
                } catch (e: ClosedReceiveChannelException) {
                    log.error { "closed receive channel " }
                    stop()
                    break
                } catch (_: CancellationException) {
                    log.error { "cancellation exception " }
                    break
                } catch (e: Exception) {
                    log.exception { "unexpected exception in TCP socket read" to e }
                    stop()
                    break
                }
            }
        }

        launch {
            val outAccess = Mutex()
            var lastSentAt = Clock.System.now()
            launch {
                while (isActive && sock.isActive) {
                    delay(500)
                    val activityTime = if(lastSentAt > lastActiveAt) lastSentAt else lastActiveAt
                    if (Clock.System.now() - activityTime > PING_INACTIVITY_TIME) {
                        log.debug { "pinging for inactivity" }
                        val repeat = outAccess.withLock {
                            try {
                                sockOutput.writeByte(0)
                                sockOutput.flush()
                                lastSentAt = Clock.System.now()
                                true
                            } catch (e: ClosedReceiveChannelException) {
                                e.printStackTrace()
                                stop()
                                false
                            } catch (_: CancellationException) {
                                false
                            } catch (e: Throwable) {
                                e.printStackTrace()
                                stop()
                                false
                            }
                        }
                        if (!repeat) break
                    }
                }
            }

            while (isActive && sock.isActive) {
                try {
                    val block = outputBlocks.receive()
                    outAccess.withLock {
                        AsyncVarint.encodeUnsigned(block.size.toULong(), sockOutput)
                        sockOutput.writeFully(block.toByteArray(), 0, block.size)
                        sockOutput.flush()
                        lastSentAt = Clock.System.now()
                    }
                } catch (_: CancellationException) {
                    log.debug { "cancellation exception on output" }
                    stop()
                    break
                } catch (_: LocalInterface.BreakConnectionException) {
                    log.debug { "requested connection break" }
                    stop()
                    break
                } catch (_: ClosedReceiveChannelException) {
                    log.debug { "receive block channel closed, closing the socket" }
                    stop()
                    break
                } catch (e: Exception) {
                    log.exception { "unexpected exception. closing." to e }
                    stop()
                    break
                }
            }
        }
    }
    val device = InetTransportDevice(inputBlocks, outputBlocks, networkAddress, {
        stop()
    })
    log.debug { "Transport ready" }
    return device
}
