package net.sergeych.kiloparsec.adapter

import io.ktor.network.sockets.*
import io.ktor.utils.io.*
import kotlinx.coroutines.channels.BufferOverflow
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.ClosedReceiveChannelException
import kotlinx.coroutines.channels.ClosedSendChannelException
import kotlinx.coroutines.coroutineScope
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import kotlinx.datetime.Clock
import net.sergeych.kiloparsec.SyncValue
import net.sergeych.mp_logger.Log
import net.sergeych.mp_logger.Loggable
import net.sergeych.mp_logger.debug
import net.sergeych.mp_logger.exception
import net.sergeych.mp_tools.globalLaunch
import kotlin.time.Duration

/**
 * This is a common part of UDP transport shared between client and server connections.
 * It should not be used directly but bu the [UdpServer], [acceptUdpDevice] and [connectUdpDevice]
 * respectively.
 */
internal class UdpSocketTransport(
    private val server: UdpConnector,
    val socketAddress: SocketAddress,
    val isServer: Boolean,
    val maxInactivityTimeout: Duration
) :
    Loggable {

    // IMPORTANT! Log stuff must be the first (or you shot your leg):
    val address = (socketAddress as InetSocketAddress).let { NetworkAddress(it.hostname, it.port) }
    override var logTag: String = "UDPT:$address${if (isServer) ":server" else ":client"}"
    override var logLevel: Log.Level? = Log.Level.DEBUG

    // Pinger params: keep them first!
    private var lastSendAt = Clock.System.now()
    private var lastReceived = Clock.System.now()
    private val pingTimeout = maxInactivityTimeout / 3
    private val pingSleep = pingTimeout / 3
    private val pingMinTimeout = pingTimeout * 2 / 3

    val inputDataBlocks = Channel<UByteArray>(256, onBufferOverflow = BufferOverflow.DROP_OLDEST)
    val outputDataBlocks = Channel<UByteArray>(256, onBufferOverflow = BufferOverflow.DROP_OLDEST)

    val inputUdpBlocks = Channel<UdpBlock>(256, onBufferOverflow = BufferOverflow.DROP_OLDEST)

    private val job = globalLaunch {
        coroutineScope {
            launch { convertOutput() }
            launch { convertInput() }
            launch { pinger() }
        }
    }

    init {
        // This is iverly important: it requires that
        // all members are initialized before use. Otherwise kotlin
        // may execute class members pr
        debug { "initialization done" }
    }


    val transportDevice: InetTransportDevice by lazy {
        InetTransportDevice(inputDataBlocks, outputDataBlocks, address, { close() }, {})
    }

    private val closedFlag = SyncValue(false)

    val isClosed: Boolean = closedFlag.value

    suspend fun close() {
        closedFlag.mutate {
            if (!it) {
                runCatching { server.sendBlock(UdpBlock.Cancel, socketAddress) }
                server.disconnectClient(socketAddress)
                runCatching { inputDataBlocks.close() }
                runCatching { outputDataBlocks.close() }
                job.cancel()
            }
            true
        }
    }


    private suspend fun send(block: UdpBlock) {
        server.sendBlock(block, socketAddress)
        lastSendAt = Clock.System.now()
    }

    /**
     * Process the block recoded by the server. Note that it should properly process all
     * block types, e.g. close on [UdpBlock.Cancel], etc. Server will not close us!
     *
     * __Important: it should not block, instead, server expects it to return ASAP__, so it
     * executes in a local coroutine context.
     *
     * Also it should not throw exceptions.
     */
    fun processIncoming(block: UdpBlock) {
        inputUdpBlocks.trySend(block)
    }

    suspend fun convertInput() {
        while(!isClosed) {
            when (val block = inputUdpBlocks.receiveCatching().getOrNull()) {

                null -> break

                is UdpBlock.Cancel -> globalLaunch {
                    debug { "received cancel block, requesting close" }
                    kotlin.runCatching { close() }
                }

                is UdpBlock.Data -> {
                    // input does not block, it uses DROP_OLDEST policy
                    lastReceived = Clock.System.now()
                    val result = kotlin.runCatching { inputDataBlocks.send(block.data) }
                    when (val e = result.exceptionOrNull()) {
                        null -> {}
                        is ClosedSendChannelException -> {
                            debug { "received close channel" }
                            close()
                        }

                        is CancellationException -> {}
                        else -> {
                            exception { "unexpected exception" to e }
                            close()
                        }
                    }
                }

                UdpBlock.Ping -> {
                    lastReceived = Clock.System.now()
                    if (lastSendAt - lastReceived > pingMinTimeout) send(UdpBlock.Ping)
                }
            }
        }
    }

    private suspend fun pinger() {
        while (!isClosed) {
            delay(pingSleep)
            val inactivity = Clock.System.now() - lastSendAt
            if( inactivity > maxInactivityTimeout) {
                debug { "inactivity timout: closing the connection" }
                close()
            }
            if (inactivity >= pingTimeout) {
                debug { "pinger sends a ping on timeout" }
                send(UdpBlock.Ping)
            }
        }
    }

    private suspend fun convertOutput() {
        while (!isClosed) {
            try {
                server.sendBlock(UdpBlock.Data(outputDataBlocks.receive()), socketAddress)
            } catch (e: CancellationException) {
                // this is ok
                break
            } catch (e: ClosedReceiveChannelException) {
                debug { "input channel is closed, closing" }
                close()
                break
            } catch (e: Exception) {
                exception { "unexpected exception in convertOutput" to e }
                close()
                break
            }
        }
        debug { "exiting convertOutput" }
    }
}